from decimal import Decimal
from io import BytesIO
from time import sleep

from flask import g

from sqlalchemy import text

from utils import match

import chellow.gas.bill_import
from chellow.models import (
    BillType,
    GContract,
    GDn,
    GReadType,
    GReadingFrequency,
    GUnit,
    Site,
    insert_bill_types,
    insert_g_read_types,
    insert_g_reading_frequencies,
    insert_g_units,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_batch_get_empty(client, sess):
    g_contract = GContract.insert_supplier(
        sess, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )

    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    sess.commit()

    response = client.get(f"/g/batches/{g_batch.id}")

    match(response, 200)


def test_batch_get(client, sess):
    site = Site.insert(sess, "22488", "Water Works")

    g_dn = GDn.insert(sess, "EE", "East of England")

    g_ldz = g_dn.insert_g_ldz(sess, "EA")

    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")

    insert_g_units(sess)

    g_unit_M3 = GUnit.get_by_code(sess, "M3")

    g_contract = GContract.insert(
        sess, False, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )

    insert_g_reading_frequencies(sess)

    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")

    g_supply = site.insert_g_supply(
        sess,
        "87614362",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )

    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    breakdown = {"units_consumed": 771}

    insert_bill_types(sess)

    bill_type_n = BillType.get_by_code(sess, "N")

    g_batch.insert_g_bill(
        sess,
        g_supply,
        bill_type_n,
        "55h883",
        "dhgh883",
        utc_datetime(2019, 4, 3),
        utc_datetime(2020, 1, 1),
        utc_datetime(2020, 1, 31, 23, 30),
        Decimal("45"),
        Decimal("12.40"),
        Decimal("1.20"),
        Decimal("14.52"),
        "",
        breakdown,
    )

    sess.commit()

    response = client.get(f"/g/batches/{g_batch.id}")

    match(response, 200)


def test_batch_add_get(client, sess):
    valid_from = to_utc(ct_datetime(2000, 1, 1))
    g_contract = GContract.insert(
        sess, False, "Fusion 2020", "", {}, valid_from, None, {}
    )

    batch_reference = "b1"
    batch_description = "Jan batch"
    g_contract.insert_g_batch(sess, batch_reference, batch_description)

    sess.commit()

    response = client.get(f"/g/supplier_contracts/{g_contract.id}/add_batch")

    match(response, 200, rf"{batch_reference}", rf"{batch_description}")


def test_batch_edit_post(sess, client):
    site = Site.insert(sess, "22488", "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    g_contract = GContract.insert(
        sess, False, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    mprn = "750278673"
    site.insert_g_supply(
        sess,
        mprn,
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )
    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")
    sess.commit()

    data = {
        "update": "Update",
        "reference": "b2",
        "description": "Feb batch",
    }

    response = client.post(f"/g/batches/{g_batch.id}/edit", data=data)

    match(response, 303, rf"/g/batches/{g_batch.id}")


def test_batch_edit_delete(sess, client):
    vf = to_utc(ct_datetime(2000, 1, 1))
    g_contract = GContract.insert(sess, False, "Fusion 2020", "", {}, vf, None, {})
    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")
    sess.commit()

    response = client.delete(f"/g/batches/{g_batch.id}/edit")

    match(response, 200)

    assert response.headers["HX-Redirect"].endswith(
        f"/g/batches?g_contract_id={g_contract.id}"
    )


def test_bill_get(client, sess):
    site = Site.insert(sess, "22488", "Water Works")

    g_dn = GDn.insert(sess, "EE", "East of England")

    g_ldz = g_dn.insert_g_ldz(sess, "EA")

    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")

    insert_g_units(sess)

    g_unit_M3 = GUnit.get_by_code(sess, "M3")

    g_contract = GContract.insert(
        sess, False, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )

    insert_g_reading_frequencies(sess)

    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")

    g_supply = site.insert_g_supply(
        sess,
        "87614362",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )

    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    breakdown = {"units_consumed": 771}

    insert_bill_types(sess)

    bill_type_n = BillType.get_by_code(sess, "N")

    g_bill = g_batch.insert_g_bill(
        sess,
        g_supply,
        bill_type_n,
        "55h883",
        "dhgh883",
        utc_datetime(2019, 4, 3),
        utc_datetime(2020, 1, 1),
        utc_datetime(2020, 1, 31, 23, 30),
        Decimal("45"),
        Decimal("12.40"),
        Decimal("1.20"),
        Decimal("14.52"),
        "",
        breakdown,
    )

    sess.commit()

    response = client.get(f"/g/bills/{g_bill.id}")

    match(response, 200, r"<tr>\s*", r"<td>units</td>\s*", r"<td>771</td>")


def test_bill_add_post(sess, client):
    site = Site.insert(sess, "22488", "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    g_contract = GContract.insert(
        sess, False, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    mprn = "750278673"
    site.insert_g_supply(
        sess,
        mprn,
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )
    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")
    insert_bill_types(sess)
    sess.commit()

    data = {
        "bill_type_id": "2",
        "mprn": mprn,
        "reference": "765988",
        "account": "1",
        "issue_year": "2017",
        "issue_month": "02",
        "issue_day": "03",
        "issue_hour": "00",
        "issue_minute": "00",
        "start_year": "2017",
        "start_month": "03",
        "start_day": "01",
        "start_hour": "00",
        "start_minute": "00",
        "finish_year": "2017",
        "finish_month": "04",
        "finish_day": "01",
        "finish_hour": "00",
        "finish_minute": "30",
        "kwh": "0.00",
        "net": "0.00",
        "vat": "0.00",
        "gross": "0.00",
        "breakdown": "{}",
    }

    response = client.post(f"/g/batches/{g_batch.id}/add_bill", data=data)

    match(response, 303, r"/g/bills/1")


def test_bill_edit_post(sess, client):
    site = Site.insert(sess, "22488", "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    g_contract = GContract.insert(
        sess, False, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    g_supply = site.insert_g_supply(
        sess,
        "87614362",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )
    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    breakdown = {"units_consumed": 771}
    insert_bill_types(sess)
    bill_type_n = BillType.get_by_code(sess, "N")
    g_bill = g_batch.insert_g_bill(
        sess,
        g_supply,
        bill_type_n,
        "55h883",
        "dhgh883",
        utc_datetime(2019, 4, 3),
        utc_datetime(2020, 1, 1),
        utc_datetime(2020, 1, 31, 23, 30),
        Decimal("45"),
        Decimal("12.40"),
        Decimal("1.20"),
        Decimal("14.52"),
        "",
        breakdown,
    )

    sess.commit()

    data = {
        "bill_type_id": "3",
        "reference": "8899900012",
        "account": "college_rooms",
        "issue_year": "2015",
        "issue_month": "11",
        "issue_day": "01",
        "issue_hour": "00",
        "issue_minute": "00",
        "start_year": "2015",
        "start_month": "09",
        "start_day": "01",
        "start_hour": "01",
        "start_minute": "00",
        "finish_year": "2015",
        "finish_month": "09",
        "finish_day": "30",
        "finish_hour": "01",
        "finish_minute": "00",
        "kwh": "4500901",
        "net_gbp": "6972.33",
        "vat_gbp": "1003.89",
        "gross_gbp": "7976.22",
        "raw_lines": "reference,mprn,bill_type,account,issue_date,"
        "start_date,finish_date,kwh,net_gbp,vat_gbp,gross_gbp,"
        "breakdown,msn,unit,correction_factor,"
        "calorific_value,prev_date,prev_value,prev_type,pres_date,"
        "prev_value,pres_type\n"
        "8899900012,750278673,N,college_rooms,2015-11-01 00:00,"
        "2015-09-01 00:00,2015-09-30 00:00,4500901,6972.33,1003.89,"
        '7976.22,{"gas_rate": 0.019448, "gas_gbp": 8936.13,'
        '"ccl_gbp": 275.32, "vat_0500pc": 0.3, "vat_1500pc": 49.12, '
        '"vat_1750pc": 55.7, "vat_2000pc": 801},hwo8tt,HCUF,FALSE,,'
        "39.300811,2015-09-01 00:00,567822,A,2015-10-01 00:00,"
        "575652,A",
        "breakdown": '{"ccl_gbp": 275.32, "gas_gbp": 8936.13, '
        '"gas_rate": 0.019448, "vat_0500pc": 0.3, "vat_1500pc": 49.12,'
        '"vat_1750pc": 55.7, "vat_2000pc": 801}',
    }

    response = client.post(f"/g/bills/{g_bill.id}/edit", data=data)

    match(response, 303, rf"/g/bills/{g_bill.id}")


def test_bill_imports_post_full(mocker, app, client, sess):
    file_lines = (
        "STX=ANA:1+Marsh Gas:MARSH Gas Limited+BPAJA:Bill Paja 771+"
        "171023:867369+856123++UTLHDR'",
        "MHD=1+UTLHDR:3'",
        "TYP=0715'",
        "SDT=Marsh Gas+Marsh Gas Limited++818671845'",
        "CDT=BPAJA:BPAJA+Bill Paja Limited - BPAJA++1'",
        "FIL=1+1+171023'",
        "MTR=6'",
        "MHD=2+UTLBIL:3'",
        "CLO=::10205041+Bill Paja Limited+Mole Hall, BATH::BA1 9MH'",
        "BCD=171022+171022+7868273476++M+N++170601:170630'",
        "CCD=1+1::GAS+MARSH6:Meter Reading++hyygk4882+87614362+170701+170601++"
        "83551:01:81773:01+8746000+771000:M3+CF:102264+841349:M3++831200:KWH+"
        "170601+170701'",
        "ADJ=1+1+CV:3930000'",
        "CCD=2+3::PPK+MARSH6:Unidentified Gas++++++++8746000++CF:100000++"
        "008521+8746000+170601+170631+008727+091'",
        "CCD=3+3::PPK+MARSH6:Commodity++++++++8746000++CF:100000++9873510+"
        "8746000+170601+170630+815510+9931'",
        "CCD=4+3::PPD+MARSH6:Transportation++++++++30000++CF:100000++86221004+"
        "30000+170601+170630+86221004+9224'",
        "CCD=5+3::PPD+MARSH6:Meter Reading++++++++30000++CF:100000++82113473+"
        "30000+170601+170630+8582284+941'",
        "CCD=6+3::PPD+MARSH6:Meter Rental++++++++30000++CF:100000++3228000+"
        "30000+170601+170630+8993000+841'",
        "CCD=7+3::PPK+MARSH6:Transportation++++++++8116500++CF:100000++"
        "005337+4617800+170601+170630+006120+882'",
        "VAT=1+++L+7986+23885+331+86334'",
        "BTL=000+88772+332++77345'",
        "MTR=14'",
        "MHD=2+UTLBIL:3'",
        "CLO=::10205041+Bill Paja Limited+Mole Hall, BATH::BA1 9MH'",
        "BCD=171022+171022+7868273476++M+N++170601:170630'",
        "CCD=1+1::GAS+MARSH6:Meter Reading++hyygk4882+87614362+170701+170601++"
        "83551:01:81773:01+8746000+771000:M3+CF:102264+841349:M3++831200:KWH+"
        "170601+170701'",
        "ADJ=1+1+CV:3930000'",
        "CCD=2+3::PPK+MARSH6:Unidentified Gas++++++++8746000++CF:100000++"
        "008521+8746000+170601+170631+008727+091'",
        "CCD=3+3::PPK+MARSH6:Commodity++++++++8746000++CF:100000++9873510+"
        "8746000+170601+170630+815510'",
        "CCD=4+3::PPD+MARSH6:Transportation++++++++30000++CF:100000++86221004+"
        "30000+170601+170630+86221004+9224'",
        "CCD=5+3::PPD+MARSH6:Meter Reading++++++++30000++CF:100000++82113473+"
        "30000+170601+170630+8582284+941'",
        "CCD=6+3::PPD+MARSH6:Meter Rental++++++++30000++CF:100000++3228000+"
        "30000+170601+170630+8993000+841'",
        "CCD=7+3::PPK+MARSH6:Transportation++++++++8116500++CF:100000++005337+"
        "4617800+170601+170630+006120+882'",
        "VAT=1+++L+7986+23885+331+86334'",
        "BTL=000+88772+332++77345'",
        "MTR=14'",
        "END=288'",
    )

    site = Site.insert(sess, "22488", "Water Works")

    g_dn = GDn.insert(sess, "EE", "East of England")

    g_ldz = g_dn.insert_g_ldz(sess, "EA")

    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")

    g_contract = GContract.insert(
        sess, False, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )

    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    insert_bill_types(sess)

    insert_g_units(sess)

    g_unit_M3 = GUnit.get_by_code(sess, "M3")

    insert_g_reading_frequencies(sess)

    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")

    site.insert_g_supply(
        sess,
        "87614362",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )

    insert_g_read_types(sess)

    sess.commit()

    file_name = "gas.engie.edi"

    file_bytes = "\n".join(file_lines).encode("utf8")

    f = BytesIO(file_bytes)

    data = {
        "g_batch_id": str(g_batch.id),
        "import_file": (f, file_name),
        "parser_name": "engie_edi",
    }

    response = client.post("/g/bill_imports", data=data)

    match(response, 303, "/g/bill_imports/0")

    importer = next(iter(chellow.gas.bill_import.importers.values()))

    for _ in range(5):
        if importer.is_alive():
            sleep(1)
        else:
            break

    response = client.get("/g/bill_imports/0")

    match(
        response,
        200,
        r"All the bills have been successfully loaded and attached to the batch.",
    )

    sess.rollback()

    res = sess.execute(text("select breakdown from g_bill where id = 2"))

    assert '"units_consumed": 771,' in next(res)[0]


def test_bill_imports_post(mocker, app, client, sess):
    g_contract = GContract.insert(
        sess,
        False,
        "Fusion 2020",
        "",
        {},
        utc_datetime(2019, 1, 1),
        utc_datetime(2019, 2, 28, 23, 30),
        {},
    )

    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    sess.commit()

    file_name = "gas.engie.edi"

    file_bytes = b"edifile'"

    f = BytesIO(file_bytes)

    parser_name = "engie_edi"
    data = {
        "g_batch_id": str(g_batch.id),
        "import_file": (f, file_name),
        "parser_name": parser_name,
    }

    import_id = 3

    mock_start_importer = mocker.patch(
        "chellow.gas.views.chellow.gas.bill_import.start_bill_importer",
        return_value=import_id,
    )

    response = client.post("/g/bill_imports", data=data)

    mock_start_importer.assert_called_with(
        g.sess, g_batch.id, file_name, file_bytes, parser_name
    )

    match(response, 303, "/g/bill_imports/3")


def test_batch_add_post(client, sess):
    g_contract = GContract.insert_supplier(
        sess,
        "Fusion 2020",
        "",
        {},
        utc_datetime(2019, 1, 1),
        utc_datetime(2019, 2, 28, 23, 30),
        {},
    )

    sess.commit()

    data = {"reference": "engie_edi", "description": "Engie EDI"}

    response = client.post(
        f"/g/supplier_contracts/{g_contract.id}/add_batch", data=data
    )

    match(response, 303, "/g/batches/1")


def test_industry_contracts(client, sess):
    contract = GContract.insert_industry(
        sess,
        "Fusion 2020",
        "",
        {},
        utc_datetime(2019, 1, 1),
        None,
        {},
    )
    contract.insert_g_rate_script(sess, utc_datetime(2019, 2, 1), {})
    sess.commit()

    response = client.get("/g/industry_contracts")

    match(response, 200, r"<td>\s*2019-01-01 00:00\s*</td>\s*<td>\s*Ongoing\s*</td>")


def test_industry_contract_edit_patch(client, sess):
    contract = GContract.insert_industry(
        sess,
        "Fusion 2020",
        "",
        {},
        utc_datetime(2019, 1, 1),
        None,
        {},
    )
    sess.commit()

    data = {
        "name": "Fusion 2020",
        "charge_script": "",
        "properties": "{}",
        "state": '{"hinderance": "sloth"}',
    }
    response = client.patch(f"/g/industry_contracts/{contract.id}/edit", data=data)
    match(response, 204)
    response = client.get(f"/g/industry_contracts/{contract.id}")
    match(response, 200, "&#34;hinderance&#34;: &#34;sloth&#34;")


def test_supplier_contracts(client, sess):
    GContract.insert_supplier(
        sess,
        "Fusion 2020",
        "",
        {},
        utc_datetime(2019, 1, 1),
        utc_datetime(2019, 2, 28, 23, 30),
        {},
    )

    sess.commit()
    response = client.get("/g/supplier_contracts")

    match(response, 200)


def test_supplier_contracts_empty(client, sess):
    response = client.get("/g/supplier_contracts")

    match(response, 200)


def test_supplies(client, sess):
    response = client.get("/g/supplies")

    match(response, 200)


def test_supply_get(client, sess):
    site = Site.insert(sess, "22488", "Water Works")

    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")

    g_contract = GContract.insert_supplier(
        sess, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )

    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")

    g_supply = site.insert_g_supply(
        sess,
        "7y94u5",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )

    sess.commit()

    response = client.get(f"/g/supplies/{g_supply.id}")

    match(response, 200)


def test_supply_note_add_get(client, sess):
    site = Site.insert(sess, "22488", "Water Works")

    g_dn = GDn.insert(sess, "EE", "East of England")

    g_ldz = g_dn.insert_g_ldz(sess, "EA")

    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")

    insert_g_units(sess)

    g_unit_M3 = GUnit.get_by_code(sess, "M3")

    g_contract = GContract.insert_supplier(
        sess, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )

    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")

    g_supply = site.insert_g_supply(
        sess,
        "7y94u5",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )

    sess.commit()

    response = client.get(f"/g/supplies/{g_supply.id}/notes/add")

    match(response, 200)


def test_supply_notes_get(client, sess):
    site = Site.insert(sess, "22488", "Water Works")

    g_dn = GDn.insert(sess, "EE", "East of England")

    g_ldz = g_dn.insert_g_ldz(sess, "EA")

    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")

    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")

    g_contract = GContract.insert_supplier(
        sess, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )

    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")

    g_supply = site.insert_g_supply(
        sess,
        "7y94u5",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )
    sess.commit()

    response = client.get(f"/g/supplies/{g_supply.id}/notes")

    match(response, 200)


def test_supply_edit_post(client, sess):
    site = Site.insert(sess, "22488", "Water Works")

    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")

    g_contract = GContract.insert_supplier(
        sess, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )

    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")

    g_supply = site.insert_g_supply(
        sess,
        "7y94u5",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )

    sess.commit()

    data = {
        "insert_g_era": "Insert",
        "start_year": "2022",
        "start_month": "01",
        "start_day": "01",
        "start_hour": "00",
        "start_minute": "00",
    }

    response = client.post(f"/g/supplies/{g_supply.id}/edit", data=data)
    assert response.headers["Location"] == f"http://localhost/g/supplies/{g_supply.id}"

    match(response, 303)


def test_read_edit_post_delete(sess, client):
    site = Site.insert(sess, "22488", "Water Works")

    g_dn = GDn.insert(sess, "EE", "East of England")

    g_ldz = g_dn.insert_g_ldz(sess, "EA")

    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")

    insert_g_units(sess)

    g_unit_M3 = GUnit.get_by_code(sess, "M3")

    g_contract = GContract.insert(
        sess, False, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )

    insert_g_reading_frequencies(sess)

    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")

    g_supply = site.insert_g_supply(
        sess,
        "7y94u5",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )

    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    insert_bill_types(sess)

    bill_type_n = BillType.get_by_code(sess, "N")

    g_bill = g_batch.insert_g_bill(
        sess,
        g_supply,
        bill_type_n,
        "55h883",
        "dhgh883",
        utc_datetime(2019, 4, 3),
        utc_datetime(2020, 1, 1),
        utc_datetime(2020, 1, 31, 23, 30),
        Decimal("45"),
        Decimal("12.40"),
        Decimal("1.20"),
        Decimal("14.52"),
        "",
        {},
    )

    insert_g_read_types(sess)

    g_read_type_A = GReadType.get_by_code(sess, "A")

    g_read = g_bill.insert_g_read(
        sess,
        "ghu5438gt",
        g_unit_M3,
        1,
        37,
        Decimal(800),
        utc_datetime(2020, 1, 1),
        g_read_type_A,
        Decimal(900),
        utc_datetime(2020, 1, 31),
        g_read_type_A,
    )

    sess.commit()

    data = {
        "delete": "Delete",
    }

    response = client.post(f"/g/reads/{g_read.id}/edit", data=data)

    match(response, 303, f"/g/bills/{g_bill.id}")


def test_supplier_rate_script_get(sess, client):
    g_contract = GContract.insert(
        sess,
        False,
        "Fusion 2020",
        "",
        {},
        utc_datetime(2019, 1, 1),
        utc_datetime(2019, 2, 28, 23, 30),
        {},
    )

    g_contract.insert_g_rate_script(sess, utc_datetime(2019, 2, 1), {})
    g_rate_script = g_contract.insert_g_rate_script(sess, utc_datetime(2019, 2, 2), {})

    sess.commit()

    response = client.get(f"/g/supplier_rate_scripts/{g_rate_script.id}")

    match(response, 200, "Previous Rate Script")


def test_supplier_rate_script_edit_post_delete(sess, client):
    g_contract = GContract.insert(
        sess,
        False,
        "Fusion 2020",
        "",
        {},
        utc_datetime(2019, 1, 1),
        utc_datetime(2019, 2, 28, 23, 30),
        {},
    )

    g_rate_script = g_contract.insert_g_rate_script(sess, utc_datetime(2019, 2, 1), {})

    sess.commit()

    data = {
        "delete": "Delete",
    }

    response = client.post(
        f"/g/supplier_rate_scripts/{g_rate_script.id}/edit", data=data
    )

    match(response, 303, f"/g/supplier_contracts/{g_contract.id}")


def test_read_add_get(sess, client):
    site = Site.insert(sess, "22488", "Water Works")

    g_dn = GDn.insert(sess, "EE", "East of England")

    g_ldz = g_dn.insert_g_ldz(sess, "EA")

    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")

    insert_g_units(sess)

    g_unit_M3 = GUnit.get_by_code(sess, "M3")

    insert_g_reading_frequencies(sess)

    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")

    g_contract = GContract.insert(
        sess, False, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )

    g_supply = site.insert_g_supply(
        sess,
        "7y94u5",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )

    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    insert_bill_types(sess)

    bill_type_n = BillType.get_by_code(sess, "N")

    g_bill = g_batch.insert_g_bill(
        sess,
        g_supply,
        bill_type_n,
        "55h883",
        "dhgh883",
        utc_datetime(2019, 4, 3),
        utc_datetime(2020, 1, 1),
        utc_datetime(2020, 1, 31, 23, 30),
        Decimal("45"),
        Decimal("12.40"),
        Decimal("1.20"),
        Decimal("14.52"),
        "",
        {},
    )

    sess.commit()

    response = client.get(f"/g/bills/{g_bill.id}/edit")

    match(response, 200)


def test_read_add_post(sess, client):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "22488", "Water Works")

    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    g_contract = GContract.insert(sess, False, "Fusion 2020", "", {}, vf, None, {})

    g_supply = site.insert_g_supply(
        sess,
        "7y94u5",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )

    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch")

    insert_bill_types(sess)

    bill_type_n = BillType.get_by_code(sess, "N")

    g_bill = g_batch.insert_g_bill(
        sess,
        g_supply,
        bill_type_n,
        "55h883",
        "dhgh883",
        utc_datetime(2019, 4, 3),
        utc_datetime(2020, 1, 1),
        utc_datetime(2020, 1, 31, 23, 30),
        Decimal("45"),
        Decimal("12.40"),
        Decimal("1.20"),
        Decimal("14.52"),
        "",
        {},
    )
    insert_g_read_types(sess)
    g_read_type_A = GReadType.get_by_code(sess, "A")

    sess.commit()

    data = {
        "msn": "fjshkk",
        "g_unit_id": g_unit_M3.id,
        "correction_factor": 1,
        "calorific_value": 1,
        "prev_date_year": 2020,
        "prev_date_month": 1,
        "prev_date_day": 4,
        "prev_date_hour": 0,
        "prev_date_minute": 0,
        "prev_value": 20,
        "prev_type_id": g_read_type_A.id,
        "pres_date_year": 2020,
        "pres_date_month": 1,
        "pres_date_day": 9,
        "pres_date_hour": 0,
        "pres_date_minute": 0,
        "pres_value": 25,
        "pres_type_id": g_read_type_A.id,
    }

    response = client.post(f"/g/bills/{g_bill.id}/add_read", data=data)

    match(response, 303)


def test_era_post_delete(sess, client):
    site = Site.insert(sess, "22488", "Water Works")

    g_dn = GDn.insert(sess, "EE", "East of England")

    g_ldz = g_dn.insert_g_ldz(sess, "EA")

    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")

    insert_g_units(sess)

    g_unit_M3 = GUnit.get_by_code(sess, "M3")

    insert_g_reading_frequencies(sess)

    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")

    g_contract = GContract.insert(
        sess, False, "Fusion 2020", "", {}, utc_datetime(2000, 1, 1), None, {}
    )

    g_supply = site.insert_g_supply(
        sess,
        "7y94u5",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
    )

    g_era = g_supply.insert_g_era_at(sess, utc_datetime(2018, 3, 1))

    sess.commit()

    data = {
        "delete": "Delete",
    }

    response = client.post(f"/g/eras/{g_era.id}/edit", data=data)

    match(response, 303, r"/g/supplies/1")



import csv
import os
import sys
import threading
import traceback
from datetime import datetime

from flask import g, request

import pytz

from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true

import chellow.dloads
from chellow.models import Era, Session, Site, SiteEra, Source, Supply
from chellow.utils import hh_after, hh_before, prev_hh, req_int
from chellow.views import chellow_redirect


def content(year, supply_id, user):
    f = writer = None
    try:
        with Session() as sess:
            running_name, finished_name = chellow.dloads.make_names(
                "crc_special_events.csv", user
            )
            f = open(running_name, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            writer.writerow(("MPAN Core", "Site Id", "Site Name", "Date", "Event"))

            year_start = datetime(year, 4, 1, tzinfo=pytz.utc)
            year_finish = prev_hh(datetime(year + 1, 4, 1, tzinfo=pytz.utc))

            def add_event(events, date, code, era=None, mpan_core=None):
                if era is None:
                    mpan_cores = [mpan_core]
                else:
                    mpan_cores = []
                    if era.imp_mpan_core is not None:
                        mpan_cores.append(era.imp_mpan_core)
                    if era.exp_mpan_core is not None:
                        mpan_cores.append(era.exp_mpan_core)

                for mpan_core in mpan_cores:
                    events.append({"date": date, "code": code, "mpan-core": mpan_core})

            if supply_id is None:
                supplies = (
                    sess.query(Supply)
                    .join(Source)
                    .join(Era)
                    .filter(
                        Source.code.in_(("net", "gen-net", "gen")),
                        Era.start_date <= year_finish,
                        or_(Era.finish_date == null(), Era.finish_date >= year_start),
                    )
                    .distinct()
                )
            else:
                supply = Supply.get_by_id(supply_id)
                supplies = sess.query(Supply).filter(Supply.id == supply.id)

            for supply in supplies:
                eras = (
                    sess.query(Era)
                    .filter(
                        Era.supply == supply,
                        Era.start_date <= year_finish,
                        or_(Era.finish_date == null(), Era.finish_date >= year_start),
                    )
                    .order_by(Era.start_date)
                    .all()
                )
                events = []
                first_era = eras[0]
                first_era_start = first_era.start_date
                if hh_after(first_era_start, year_start):
                    add_event(events, first_era_start, "New Supply", first_era)

                last_era = eras[-1]
                last_era_finish = last_era.finish_date
                if hh_before(last_era_finish, year_finish):
                    add_event(events, last_era_finish, "Disconnection", last_era)

                prev_era = first_era
                for era in eras[1:]:
                    if era.msn != prev_era.msn:
                        add_event(events, era.start_date, "Meter Change", era)
                    if era.pc.code != prev_era.pc.code:
                        add_event(
                            events, era.start_date, "Change Of Profile Class", era
                        )

                    if era.mop_contract_id != prev_era.mop_contract_id:
                        add_event(events, era.start_date, "Change Of MOP", era)

                    if era.dc_contract_id != prev_era.dc_contract_id:
                        add_event(events, era.start_date, "Change Of DC", era)

                    for is_import in [True, False]:
                        if era.imp_mpan_core is None:
                            mpan_core = era.exp_mpan_core
                        else:
                            mpan_core = era.imp_mpan_core

                        if is_import:
                            cur_sup = era.imp_supplier_contract
                            prev_sup = prev_era.imp_supplier_contract
                        else:
                            cur_sup = era.exp_supplier_contract
                            prev_sup = prev_era.exp_supplier_contract

                        if cur_sup is None and prev_sup is not None:
                            add_event(
                                events, era.start_date, "End of supply", mpan_core
                            )
                        elif cur_sup is not None and prev_sup is None:
                            add_event(
                                events,
                                era.start_date,
                                "Start of supply",
                                None,
                                mpan_core,
                            )
                        elif (
                            cur_sup is not None
                            and prev_sup is not None
                            and cur_sup != prev_sup
                        ):
                            add_event(
                                events,
                                era.start_date,
                                "Change Of Supplier",
                                None,
                                mpan_core,
                            )

                    prev_era = era

                if len(events) > 0:
                    site = (
                        sess.query(Site)
                        .join(SiteEra)
                        .filter(SiteEra.is_physical == true(), SiteEra.era == last_era)
                        .one()
                    )

                    for event in events:
                        vals = [
                            event["mpan-core"],
                            site.code,
                            site.name,
                            event["date"].strftime("%Y-%m-%d %H:%M"),
                            event["code"],
                        ]
                        writer.writerow(vals)

                # Avoid a long-running transaction
                sess.rollback()
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    year = req_int("year")
    supply_id = req_int("supply_id") if "supply_id" in request.values else None

    threading.Thread(target=content, args=(year, supply_id, g.user)).start()
    return chellow_redirect("/downloads", 303)



def ccl(data_source):
    try:
        ccl_cache = data_source.caches["ccl"]
    except KeyError:
        ccl_cache = data_source.caches["ccl"] = {}

    for hh in data_source.hh_data:
        try:
            rate = ccl_cache[hh["start-date"]]
        except KeyError:
            hh_start = hh["start-date"]
            rates = data_source.non_core_rate("ccl", hh_start)
            rate = ccl_cache[hh_start] = float(rates["ccl_gbp_per_msp_kwh"])

        hh["ccl-kwh"] = hh["msp-kwh"]
        hh["ccl-rate"] = rate
        hh["ccl-gbp"] = hh["msp-kwh"] * rate



from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from dateutil.relativedelta import relativedelta

from openpyxl import load_workbook

from werkzeug.exceptions import BadRequest

from chellow.models import Session
from chellow.utils import hh_format, parse_mpan_core, to_utc


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f, data_only=True)
        self.sheet = self.book.worksheets[0]

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_number = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def get_ct_date(self, col, row):
        cell = self.get_cell(col, row)
        val = cell.value
        if not isinstance(val, Datetime):
            raise BadRequest(
                f"Problem reading {val} as a timestamp at {cell.coordinate}."
            )
        return val

    def get_start_date(self, col, row):
        dt = self.get_ct_date(col, row)
        return None if dt is None else to_utc(dt)

    def get_cell(self, col, row):
        try:
            coordinates = f"{col}{row}"
            return self.sheet[coordinates]
        except IndexError:
            raise BadRequest(
                f"Can't find the cell {coordinates} on sheet {self.sheet}."
            )

    def get_str(self, col, row):
        return self.get_cell(col, row).value.strip()

    def get_dec(self, col, row):
        cell = self.get_cell(col, row)
        try:
            return Decimal(str(cell.value))
        except InvalidOperation as e:
            raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")

    def get_int(self, col, row):
        return int(self.get_cell(col, row).value)

    def make_raw_bills(self):
        row_index = None
        try:
            with Session() as sess:
                bills = []
                issue_date_str = self.get_str("A", 7)
                issue_date = Datetime.strptime(issue_date_str[6:], "%d/%m/%Y %H:%M:%S")
                for row in range(12, len(self.sheet["A"]) + 1):
                    val = self.get_cell("A", row).value
                    if val is None or val == "":
                        break

                    self._set_last_line(row_index, val)
                    mpan_core = parse_mpan_core(str(self.get_int("B", row)))
                    start_date = self.get_start_date("D", row)
                    finish_date = self.get_start_date("E", row) + relativedelta(
                        hours=23, minutes=30
                    )

                    net = round(self.get_dec("W", row), 2)

                    cop_3_meters = self.get_int("G", row)
                    cop_3_rate = self.get_dec("H", row)
                    cop_3_gbp = self.get_dec("I", row)

                    # Cop 5 meters
                    self.get_int("J", row)
                    cop_5_rate = self.get_dec("K", row)
                    cop_5_gbp = self.get_dec("L", row)

                    ad_hoc_visits = self.get_dec("P", row)
                    ad_hoc_rate = self.get_dec("Q", row)
                    ad_hoc_gbp = self.get_dec("R", row)

                    annual_visits = self.get_int("S", row)
                    annual_rate = self.get_dec("T", row)
                    annual_gbp = self.get_dec("U", row)

                    if cop_3_meters > 0:
                        cop = "3"
                        mpan_rate = cop_3_rate
                        mpan_gbp = cop_3_gbp
                    else:
                        cop = "5"
                        mpan_rate = cop_5_rate
                        mpan_gbp = cop_5_gbp

                    breakdown = {
                        "raw_lines": [],
                        "cop": [cop],
                        "settlement-status": ["settlement"],
                        "mpan-rate": [mpan_rate],
                        "mpan-gbp": mpan_gbp,
                        "ad-hoc-visits": ad_hoc_visits,
                        "ad-hoc-rate": [ad_hoc_rate],
                        "ad-hoc-gbp": ad_hoc_gbp,
                        "annual-visits-count": annual_visits,
                        "annual-visits-rate": [annual_rate],
                        "annual-visits-gbp": annual_gbp,
                    }
                    annual_date_cell = self.get_cell("V", row)
                    annual_date_value = annual_date_cell.value
                    if annual_date_value is not None:
                        if isinstance(annual_date_value, Datetime):
                            annual_date = hh_format(annual_date_value)
                        else:
                            annual_date = annual_date_value
                        breakdown["annual-visits-date"] = [annual_date]

                    bills.append(
                        {
                            "bill_type_code": "N",
                            "kwh": Decimal(0),
                            "vat": Decimal("0.00"),
                            "net": net,
                            "gross": net,
                            "reads": [],
                            "breakdown": breakdown,
                            "account": mpan_core,
                            "issue_date": issue_date,
                            "start_date": start_date,
                            "finish_date": finish_date,
                            "mpan_core": mpan_core,
                            "reference": "_".join(
                                (
                                    start_date.strftime("%Y%m%d"),
                                    finish_date.strftime("%Y%m%d"),
                                    issue_date.strftime("%Y%m%d"),
                                    mpan_core,
                                )
                            ),
                        }
                    )
                    sess.rollback()

        except BadRequest as e:
            raise BadRequest(f"Row number: {row} {e.description}")

        return bills



import atexit
import collections
import threading
import traceback
from decimal import Decimal

import requests

from werkzeug.exceptions import BadRequest

from zish import loads

from chellow.models import Contract, RateScript, Session
from chellow.utils import (
    HH,
    c_months_u,
    ct_datetime,
    hh_format,
    to_ct,
    to_utc,
    utc_datetime_now,
)


def hh(data_source, provider="APXMIDP"):
    try:
        cache = data_source.caches["bmarketidx"]
    except KeyError:
        cache = data_source.caches["bmarketidx"] = {}

    for hh in data_source.hh_data:
        try:
            hh["bmarketidx-rate"] = cache[hh["start-date"]][provider]
        except KeyError:
            h_start = hh["start-date"]
            rates = data_source.non_core_rate("bmarketidx", h_start)["rates"]

            try:
                idxs = cache[h_start]
            except KeyError:
                idxs = cache[h_start] = {}

            try:
                rate = rates[h_start]
            except KeyError:
                try:
                    rate = sorted(rates.items())[-1][1]
                except KeyError:
                    raise BadRequest(
                        f"For the bmarketidx rate script at "
                        f"{hh_format(h_start)} the rate cannot be found."
                    )

            try:
                idx = rate[provider]
            except KeyError:
                try:
                    idx = sorted(rate.items())[0][1]
                except KeyError:
                    raise BadRequest(
                        f"For the bmarketidx rate script at "
                        f"{hh_format(h_start)} a rate cannot be found for the "
                        f"provider {provider}."
                    )

            hh["bmarketidx-rate"] = idxs[provider] = float(idx)


bmarketidx_importer = None


class BmarketidxImporter(threading.Thread):
    def __init__(self):
        super().__init__(name="Bmarketidx Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=100)
        self.stopped = threading.Event()
        self.going = threading.Event()

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message
        )

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                self.global_alert = contract = None
                with Session() as sess:
                    try:
                        self.log("Starting to check bmarketidx.")
                        contract = Contract.get_non_core_by_name(sess, "bmarketidx")
                        latest_rs = (
                            sess.query(RateScript)
                            .filter(RateScript.contract_id == contract.id)
                            .order_by(RateScript.start_date.desc())
                            .first()
                        )
                        start_ct = to_ct(latest_rs.start_date)

                        months = list(
                            c_months_u(
                                start_year=start_ct.year,
                                start_month=start_ct.month,
                                months=2,
                            )
                        )
                        month_start, month_finish = months[1]

                        now = utc_datetime_now()
                        if now > month_finish:
                            _process_month(
                                self.log,
                                sess,
                                contract,
                                latest_rs,
                                month_start,
                                month_finish,
                            )

                    except BaseException:
                        self.log(f"Outer problem {traceback.format_exc()}")
                        sess.rollback()
                        if contract is None:
                            self.global_alert = (
                                "There's a problem with the bmarketidx automatic "
                                "importer."
                            )
                        else:
                            self.global_alert = (
                                f"There's a problem with the <a "
                                f"href='/e/non_core_contracts/{contract.id}/"
                                f"automatic_importer'>bmarketidx automatic "
                                f"importer</a>."
                            )
                    finally:
                        self.lock.release()
                        self.log("Finished checking bmarketidx rates.")

            self.going.wait(2 * 60 * 60)
            self.going.clear()


def _process_month(log_f, sess, contract, latest_rs, month_start, month_finish):
    latest_rs_id = latest_rs.id
    log_f(
        f"Checking to see if data is available from {hh_format(month_start)} "
        f"to {hh_format(month_finish)} on BMRS."
    )
    rates = {}
    month_finish_ct = to_ct(month_finish)
    for d in range(month_finish_ct.day):
        day_ct = ct_datetime(month_finish_ct.year, month_finish_ct.month, d + 1)
        params = {
            "q": f"ajax/alldata/MID/Date,SP,Provider,Price,Volume/NULL/"
            f'{day_ct.strftime("%Y-%m-%d")}/ALL'
        }
        sess.rollback()
        r = requests.get(
            "https://www.bmreports.com/bmrs/", params=params, timeout=60, verify=False
        )
        res = r.json()
        for h in res["arr"]:
            dt = to_utc(day_ct + (int(h["settlementPeriod"]) - 1) * HH)
            try:
                rate = rates[dt]
            except KeyError:
                rate = rates[dt] = {}
            rate[h["DataProviderId"]] = Decimal(h["MarketIndexPrice"]) / Decimal(1000)

    if month_finish in rates:
        log_f("The whole month's data is there.")
        script = {"rates": rates}
        rs = RateScript.get_by_id(sess, latest_rs_id)
        contract.update_rate_script(
            sess, rs, rs.start_date, month_finish, loads(rs.script)
        )
        contract.insert_rate_script(sess, month_start, script)
        sess.commit()
        log_f(f"Added a new rate script starting at {hh_format(month_start)}.")
    else:
        msg = "There isn't a whole month there yet."
        if len(rates) > 0:
            msg += " The last date is {sorted(rates.keys())[-1]}"
        log_f(msg)


def get_importer():
    return bmarketidx_importer


def startup():
    global bmarketidx_importer
    bmarketidx_importer = BmarketidxImporter()
    bmarketidx_importer.start()


@atexit.register
def shutdown():
    global bmarketidx_importer

    if bmarketidx_importer is not None:
        bmarketidx_importer.stop()



import csv
import decimal
from decimal import Decimal

from werkzeug.exceptions import BadRequest

from xlrd import open_workbook

from chellow.utils import HH, utc_datetime, utc_datetime_now


def get_value(row, title_row, name):
    try:
        idx = title_row.index(name)
    except ValueError:
        return None

    try:
        val = row[idx].value
    except IndexError:
        raise BadRequest(
            f"For the name '{name}', the index is {idx} which is beyond the "
            f"end of the row."
        )

    if isinstance(val, str):
        return val.strip()
    else:
        return val


def get_dec(row, title_row, name):
    val = get_value(row, title_row, name)

    try:
        return val if val is None else Decimal(str(val))
    except decimal.InvalidOperation:
        return None


def get_rate(row, title_row, name):
    raw = get_dec(row, title_row, name)
    return None if raw is None else [raw / Decimal("100")]


def get_date_ct(row, title_row, name):
    s = get_value(row, title_row, name)
    nums = list(map(int, s.split("-")))
    return utc_datetime(nums[0], nums[1], nums[2])


def _parse_row(row, row_index, datemode, title_row):
    titles = [c.value for c in title_row]
    bill_start_date = get_date_ct(row, titles, "BILL START DATE")
    bill_finish_date = get_date_ct(row, titles, "BILL END DATE") + (HH * 47)
    issue_date = utc_datetime_now()

    aahedc_kwh = get_dec(row, titles, "LVY-AAHEDC-ALL USE KWH")

    vals = [
        ("capmechob-kwh", get_dec(row, titles, "LVY-CMLOB-ALL USE KWH")),
        ("capmechob-rate", get_dec(row, titles, "LVY-CMLOB-ALL RATE P/KWH")),
        ("capmechob-gbp", get_dec(row, titles, "LVY-CMLOB-ALL COST GBP")),
        ("capmechop-kwh", get_dec(row, titles, "LVY-CMLOP-ALL USE KWH")),
        ("capmechop-rate", get_dec(row, titles, "LVY-CMLOP-ALL RATE P/KWH")),
        ("capmechop-gbp", get_dec(row, titles, "LVY-CMLOP-ALL COST GBP")),
        ("duos-fixed-days", get_dec(row, titles, "DUOS-STND-SITE USE DAY")),
        (
            "duos-availability-kva-days",
            [get_dec(row, titles, "DUOS-AVAIL-AV USE KVADAY")],
        ),
        (
            "duos-availability-rate",
            get_rate(row, titles, "DUOS-AVAIL-AV RATE P/KVADAY"),
        ),
        ("duos-availability-gbp", get_dec(row, titles, "DUOS-AVAIL-AV COST GBP")),
        ("duos-fixed-rate", get_rate(row, titles, "DUOS-STND-SITE RATE P/DAY")),
        ("duos-fixed-gbp", get_dec(row, titles, "DUOS-STND-SITE COST GBP")),
        ("duos-amber-kwh", get_dec(row, titles, "DUOS-UNIT-AMBER USE KWH")),
        ("duos-amber-rate", get_rate(row, titles, "DUOS-UNIT-AMBER RATE P/KWH")),
        ("duos-amber-gbp", get_dec(row, titles, "DUOS-UNIT-AMBER COST GBP")),
        ("duos-green-kwh", get_dec(row, titles, "DUOS-UNIT-GREEN USE KWH")),
        ("duos-green-rate", get_rate(row, titles, "DUOS-UNIT-GREEN RATE P/KWH")),
        ("duos-green-gbp", get_dec(row, titles, "DUOS-UNIT-GREEN COST GBP")),
        ("duos-red-kwh", get_dec(row, titles, "DUOS-UNIT-RED USE KWH")),
        ("duos-red-rate", get_rate(row, titles, "DUOS-UNIT-RED RATE P/KWH")),
        ("duos-red-gbp", get_dec(row, titles, "DUOS-UNIT-RED COST GBP")),
        ("aahedc-kwh", aahedc_kwh),
        ("aahedc-rate", get_rate(row, titles, "LVY-AAHEDC-ALL RATE P/KWH")),
        ("aahedc-gbp", get_dec(row, titles, "LVY-AAHEDC-ALL COST GBP")),
        ("bsuos-nbp-kwh", get_dec(row, titles, "LVY-BSUOS-ALL USE KWH")),
        ("bsuos-rate", get_rate(row, titles, "LVY-BSUOS-ALL RATE P/KWH")),
        ("bsuos-gbp", get_dec(row, titles, "LVY-BSUOS-ALL COST GBP")),
        ("ccl-kwh", get_dec(row, titles, "LVY-CCL-ALL USE KWH")),
        ("ccl-rate", get_rate(row, titles, "LVY-CCL-ALL RATE P/KWH")),
        ("ccl-gbp", get_dec(row, titles, "LVY-CCL-ALL COST GBP")),
        ("cfdob-kwh", get_dec(row, titles, "LVY-CFDOB-ALL USE KWH")),
        ("cfdob-rate", get_rate(row, titles, "LVY-CFDOB-ALL RATE P/KWH")),
        ("cfdob-gbp", get_dec(row, titles, "LVY-CFDOB-ALL COST GBP")),
        ("cfdop-kwh", get_dec(row, titles, "LVY-CFDOP-ALL USE KWH")),
        ("cfdop-rate", get_rate(row, titles, "LVY-CFDOP-ALL RATE P/KWH")),
        ("cfdop-gbp", get_dec(row, titles, "LVY-CFDOP-ALL COST GBP")),
        ("fit-kwh", get_dec(row, titles, "LVY-FIT-ALL USE KWH")),
        ("fit-rate", get_rate(row, titles, "LVY-FIT-ALL RATE P/KWH")),
        ("fit-gbp", get_dec(row, titles, "LVY-FIT-ALL COST GBP")),
        ("ro-kwh", get_dec(row, titles, "LVY-RO-ALL USE KWH")),
        ("ro-rate", get_rate(row, titles, "LVY-RO-ALL RATE P/KWH")),
        ("ro-gbp", get_dec(row, titles, "LVY-RO-ALL COST GBP")),
        ("summer-kwh", get_dec(row, titles, "NRG-UNIT-SUMMER USE KWH")),
        ("summer-rate", get_rate(row, titles, "NRG-UNIT-SUMMER RATE P/KWH")),
        ("summer-gbp", get_dec(row, titles, "NRG-UNIT-SUMMER COST GBP")),
        ("winter-kwh", get_dec(row, titles, "NRG-UNIT-WINTER USE KWH")),
        ("winter-rate", get_rate(row, titles, "NRG-UNIT-WINTER RATE P/KWH")),
        ("winter-gbp", get_dec(row, titles, "NRG-UNIT-WINTER COST GBP")),
        ("admin-months", get_dec(row, titles, "SYS-ADMIN-ALL USE MO")),
        ("admin-rate", [get_dec(row, titles, "SYS-ADMIN-ALL RATE GBP/MO")]),
        ("admin-gbp", get_dec(row, titles, "SYS-ADMIN-ALL COST GBP")),
        ("triad-days", get_dec(row, titles, "TUOS-TRIAD-SITE USE DAY")),
        ("triad-rate", [get_dec(row, titles, "TUOS-TRIAD-SITE RATE GBP/DAY")]),
        ("triad-gbp", get_dec(row, titles, "TUOS-TRIAD-SITE COST GBP")),
    ]

    bd = {
        "raw_lines": [str(title_row), str(row)],
    }

    for k, v in vals:
        if v is not None and v != [None]:
            bd[k] = v

    kwh = Decimal("0.00") if aahedc_kwh is None else aahedc_kwh
    net_gbp = Decimal("0.00")
    net_gbp += sum(v for k, v in bd.items() if k.endswith("-gbp"))
    vat_gbp = Decimal("0.00")
    reference = issue_date.strftime("%Y%m%dT%H%M") + "_" + str(row_index + 1)

    return {
        "bill_type_code": "N",
        "kwh": kwh,
        "vat": Decimal("0.00"),
        "net": net_gbp,
        "gross": net_gbp + vat_gbp,
        "reads": [],
        "breakdown": bd,
        "account": "2007",
        "issue_date": issue_date,
        "start_date": bill_start_date,
        "finish_date": bill_finish_date,
        "mpan_core": "22 0003 0354 632",
        "reference": reference,
    }


class Parser:
    def __init__(self, f):
        self.book = open_workbook(file_contents=f.read())
        self.sheet = self.book.sheet_by_index(1)

        self.last_line = None
        lines = (self._set_last_line(i, l) for i, l in enumerate(f))
        self.reader = csv.reader(lines, skipinitialspace=True)
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        bills = []
        title_row = self.sheet.row(0)
        for row_index in range(1, self.sheet.nrows):
            row = self.sheet.row(row_index)
            datemode = self.book.datemode
            try:
                bills.append(_parse_row(row, row_index, datemode, title_row))
            except BadRequest as e:
                raise BadRequest("On row " + str(row_index) + ": " + str(e.description))
        return bills



from datetime import datetime as Datetime
from decimal import Decimal
from io import BytesIO

from openpyxl import Workbook

from chellow.e.bill_parsers.nonsettlement_dc_stark_xlsx import Parser
from chellow.utils import utc_datetime


def test(sess):
    with open(
        "test/e/bill_parsers/test_bill_parser_nonsettlement_dc_stark_xlsx/"
        "bills.nonsettlement.dc.stark.xlsx",
        mode="rb",
    ) as f:
        parser = Parser(f)
        parser.make_raw_bills()


def test_old(sess):
    with open(
        "test/e/bill_parsers/test_bill_parser_nonsettlement_dc_stark_xlsx/"
        "bills_old.nonsettlement.dc.stark.xlsx",
        mode="rb",
    ) as f:
        parser = Parser(f)
        parser.make_raw_bills()


def test_one_bill(sess):
    f = BytesIO()
    wb = Workbook()
    sheet = wb.worksheets[0]
    sheet.insert_rows(0, 2)
    sheet.insert_cols(0, 9)

    sheet["A1"] = "NAME"
    sheet["B1"] = "NMR"
    sheet["C1"] = "METER TYPE"
    sheet["D1"] = "IDENTIFIER"
    sheet["E1"] = "METER"
    sheet["F1"] = "MPAN REF"
    sheet["G1"] = "START"
    sheet["H1"] = "END"
    sheet["I1"] = "CHECK"

    sheet["A2"] = "Quantum Computer"
    sheet["B2"] = 65837766
    sheet["C2"] = "x600"
    sheet["D2"] = 7493477439
    sheet["E2"] = 7864739737
    sheet["F2"] = 2088757371777
    sheet["G2"] = Datetime(2022, 2, 1)
    sheet["H2"] = Datetime(2022, 2, 28)
    sheet["I2"] = "Billed"

    sheet["C6"].value = Datetime(2022, 3, 1)

    sheet["B12"].value = "2227651294270"
    sheet["C12"].value = "CS"
    sheet["D12"].value = "Settled"
    sheet["E12"].value = "Massive Field, Taunton, Somerset"
    sheet["F12"].value = Datetime(2022, 1, 1)
    sheet["G12"].value = Datetime(2022, 1, 31)
    sheet["H12"].value = "5.00"
    sheet["I12"].value = "0.12"
    sheet["J12"].value = "0.76"
    sheet["K12"].value = "0.99"

    wb.save(f)
    f.seek(0)
    p = Parser(f)
    result = p.make_raw_bills()

    expected = [
        {
            "bill_type_code": "N",
            "kwh": Decimal("0"),
            "vat": Decimal("1.00"),
            "net": Decimal("5.00"),
            "gross": Decimal("6.00"),
            "reads": [],
            "breakdown": {
                "raw_lines": [],
                "cop": ["5"],
                "settlement-status": ["non_settlement"],
                "msn": ["7864739737"],
                "meter-rate": [Decimal("60.00")],
                "meter-gbp": Decimal("5.00"),
            },
            "account": "20 8875 7371 777",
            "issue_date": utc_datetime(2022, 2, 1),
            "start_date": utc_datetime(2022, 2, 1),
            "finish_date": utc_datetime(2022, 2, 28, 23, 30),
            "mpan_core": "20 8875 7371 777",
            "reference": "20220201_20220228_20220201_20 8875 7371 777",
        }
    ]

    assert result == expected
